'''QR Authentication'''
import logging

from px_device_identity import DeviceProperties, Device
from requests import ConnectionError, get

log = logging.getLogger(__name__)


class QRAuthentication:
    '''QR Authentication'''

    def __init__(self, device: Device):
        self.device_properties: DeviceProperties = device.properties
        self.access_token: str = device.get_access_token()['access_token']

    def _make_request(self, url: str, access_token: str):
        '''Initiate API request'''
        log.debug('=> Making request to {}'.format(url))
        try:
            res = get(url, headers={
                'authorization': 'Bearer {}'.format(access_token)
            })
            res.raise_for_status()
            return res.json()
        except ConnectionError as err:
            log.error("Connection to {} failed.".format(url))
            raise err
        except Exception as err:
            log.error(err)
            raise err

    def login(self):
        '''Request user QR authentication grant'''
        log.debug('=> Getting user auth grant')
        url = f"{self.device_properties.host}/auth/qr"
        data = self._make_request(url, self.access_token)
        # auth_req_id, exp
        return data

    def status(self, auth_req_id: str):
        '''Request user QR authentication grant status'''
        log.debug('=> Getting user auth request status')
        url = f"{self.device_properties.host}/auth/qr/{auth_req_id}"
        data = self._make_request(url, self.access_token)
        # status, access_token, issue_time
        return data



'''Common Authentication'''
import logging

from px_device_identity import Device, DeviceProperties
from requests import ConnectionError, post

log = logging.getLogger(__name__)


class CommonAuthentication:
    '''Common Authentication'''

    def __init__(self, device: Device):
        self.device_properties: DeviceProperties = device.properties
        self.get_device_jwt = device.get_device_jwt


    def refresh(self, refresh_token: str, grant_type: str = 'refresh_token'):
        '''Refresh token'''
        log.debug('=> Requesting new access- and refresh-token.')
        url = f"{self.device_properties.host}/oidc/token"
        device_jwt = self.get_device_jwt()
        req_content = {
            "grant_type": grant_type,
            "client_id": self.device_properties.client_id,
            "client_assertion_type": "urn:ietf:params:oauth:client-assertion-type:jwt-bearer",
            "client_assertion": device_jwt['device_jwt'],
            "refresh_token": refresh_token
        }

        try:
            res = post(url, data=req_content)
            res.raise_for_status()
            return res.json()
        except ConnectionError as err:
            log.error("Connection to {} failed.".format(url))
            raise err
        except Exception as err:
            log.error(err)
            raise err



'''CIBA Authentication'''
import logging
from uuid import uuid4

from px_device_identity import Device, DeviceProperties
from px_device_identity.device.jwt import (
    device_singing_base, get_unix_time_in_seconds
)
from px_device_identity.device.sign import Sign
from requests import ConnectionError, post
from typing import Union

log = logging.getLogger(__name__)


class CIBAAuthentication():
    '''CIBA Authentication'''

    def __init__(self, device: Device):
        self.device_properties: 'DeviceProperties' = device.properties
        self.device_jwt = device.get_device_jwt()['device_jwt']
        self.config_dir = device.config_dir
        self.key_dir = device.key_dir

    def _auth_request(self, login_hint: str, is_login_hint_token: bool, message: str, scope: str, resource: Union[None, str] = None):
        iat = get_unix_time_in_seconds()
        exp = iat + 300
        initial_request = {
            'scope': scope,
            'binding_message': message,
            'iss': self.device_properties.client_id,
            'aud': f"{self.device_properties.host}/oidc",
            'exp': exp,
            'iat': iat,
            'nbf': iat - 10,
            'jti': uuid4().hex,
        }
        
        if is_login_hint_token:
            initial_request['login_hint_token'] = login_hint
        else:
            initial_request['login_hint'] = login_hint
            
        if resource:
            initial_request['resource'] = resource
        
        signature_content = device_singing_base(
            initial_request,
            self.device_properties.client_id,
            self.device_properties.host + '/oidc',
            iat,
            exp
        )
        signature = Sign(
            self.device_properties,
            signature_content,
            self.key_dir,
        ).sign()
        result = f"{signature_content}.{signature}"
        return result

    def _auth_request_content(self, login_hint: str, is_login_hint_token: bool, message: str, scope: str, resource: Union[None, str] = None):
        '''CIBA auth request content'''
        auth_request = {
            "request": self._auth_request(
                login_hint,
                is_login_hint_token,
                message,
                scope,
                resource,
            ),
            "client_id": self.device_properties.client_id,
            "client_assertion_type": "urn:ietf:params:oauth:client-assertion-type:jwt-bearer",
            "client_assertion": self.device_jwt,
        }
        return auth_request

    def _status_request_content(self, auth_req_id: str):
        '''CIBA status request content'''
        status_request = {
            'grant_type': 'urn:openid:params:grant-type:ciba',
            'auth_req_id': auth_req_id,
            'client_id': self.device_properties.client_id,
            'client_assertion_type': 'urn:ietf:params:oauth:client-assertion-type:jwt-bearer',
            'client_assertion': self.device_jwt
        }
        return status_request

    def _make_request(self, url: str, data: dict):
        '''Make API request'''
        try:
            res = post(url, data=data, headers={
                'Accept': 'application/json, text/plain, */*',
                'Content-Type': 'application/x-www-form-urlencoded'
            })
            res.raise_for_status()
            return res.json()
        except ConnectionError as err:
            log.error(f"Connection to {url} failed.")
            raise err
        except Exception as err:
            log.error(err)
            raise err

    def login(self, login_hint: str, is_login_hint_token: bool, message: str = 'Authorize Login', scope: str = "openid contact profile", resource: Union[None, str] = None):
        '''
        CIBA (Backchannel) login request

        Params:
            login_hint_token: username
            message: Optional login message
                     Recommended format 'Authorize login to ApplicationName'
        '''
        log.debug(f"=> CIBA login request for {login_hint}")
        url = self.device_properties.host + '/oidc/backchannel'
        return self._make_request(url, self._auth_request_content(
            login_hint,
            is_login_hint_token,
            message,
            scope,
            resource
        ))

    def status(self, auth_req_id: str):
        '''CIBA (Backchannel) login status request'''
        log.debug(f"=> CIBA login status request for ID {auth_req_id}")
        url = self.device_properties.host + '/oidc/token'
        return self._make_request(url, self._status_request_content(auth_req_id))



'''PantherX User Identity Service'''

import logging
import platform

import pkg_resources
from flask import Flask, request, jsonify, make_response, g
from px_device_identity import Device, is_superuser_or_quit
from requests.models import HTTPError
from werkzeug.middleware.proxy_fix import ProxyFix
from waitress import serve

from .common import CommonAuthentication
from .ciba import CIBAAuthentication
from .cli import get_cl_arguments
from .log import *
from .qr import QRAuthentication

log = logging.getLogger(__name__)

opsys = platform.system()
version = pkg_resources.require("px-user-identity-service")[0].version

app = Flask(__name__)
app.wsgi_app = ProxyFix(app.wsgi_app)

@app.before_request
def before_request():
    g.key_dir = app.config['key_dir']
    g.config_dir = app.config['config_dir']
    g.debug = app.config['debug']

@app.after_request
def cors_headers(response):
    response.headers['Access-Control-Allow-Origin'] = '127.0.0.1'
    if request.method == 'OPTIONS':
        response.headers['Access-Control-Allow-Methods'] = response.headers.get('Allow', '127.0.0.1')
        response.headers['Access-Control-Allow-Headers'] = request.headers.get('Access-Control-Request-Headers', '127.0.0.1')
        response.headers['Access-Control-Max-Age'] = '86400'  # 24 hours
    return response

@app.route('/auth/qr', methods=['GET', 'OPTIONS'])
def qr_auth_request():
    if request.method == 'OPTIONS':
        return make_response('', 204)

    device = Device(
        key_dir=g.key_dir,
        config_dir=g.config_dir,
    )

    try:
        data = QRAuthentication(device).login()
        return make_response(jsonify(data), 202)
    except Exception as err:
        log.error(err)
        return make_response(jsonify(title='Client error', description=str(err)), 400)

@app.route('/auth/qr/<auth_req_id>', methods=['GET', 'OPTIONS'])
def qr_auth_status(auth_req_id):
    if request.method == 'OPTIONS':
        return make_response('', 204)

    device = Device(
        key_dir=g.key_dir,
        config_dir=g.config_dir,
    )

    try:
        data = QRAuthentication(device).status(auth_req_id)
        return make_response(jsonify(data), 202)
    except Exception as err:
        log.error(err)
        return make_response(jsonify(title='Client error', description=str(err)), 400)

@app.route('/auth/refresh', methods=['POST', 'OPTIONS'])
def common_auth_refresh():
    if request.method == 'OPTIONS':
        return make_response('', 204)

    access_token = request.json.get('access_token', None)
    refresh_token = request.json.get('refresh_token', None)
    grant_type = 'refresh_token'

    if access_token is None or refresh_token is None:
        return make_response(jsonify(title='Bad request', description='The access_token and refresh_token are required.'), 400)

    device = Device(
        key_dir=g.key_dir,
        config_dir=g.config_dir,
    )
    auth = CommonAuthentication(device)

    try:
        data = auth.refresh(refresh_token, grant_type)
        return make_response(jsonify(data), 200)
    except Exception as err:
        log.error(err)
        return make_response(jsonify(title='Client error', description='Could not refresh token.'), 400)

@app.route('/auth/bc', methods=['POST', 'OPTIONS'])
def bc_auth_request():
    if request.method == 'OPTIONS':
        return make_response('', 204)

    scope = request.json.get('scope', "openid contact profile")
    login_hint = request.json.get('login_hint', None)
    login_hint_token = request.json.get('login_hint_token', None)
    login_message = request.json.get('login_message', None)
    resource = request.json.get('resource', None)

    if login_hint_token is None and login_hint is None:
        return make_response(jsonify(title='Bad request', description='The login_hint_token or login_hint is required.'), 400)

    if login_message is None:
        return make_response(jsonify(title='Bad request', description='The login_message is required.'), 400)
    
    is_login_hint_token = False
    if login_hint_token is not None:
        is_login_hint_token = True
        login_hint = login_hint_token
    else:
        is_login_hint_token = False
        login_hint = login_hint

    device = Device(
        key_dir=g.key_dir,
        config_dir=g.config_dir,
    )
    ciba = CIBAAuthentication(device)

    try:
        data = data = ciba.login(
            login_hint,
            is_login_hint_token,
            login_message,
            scope,
            resource
        )
        return make_response(jsonify(data), 202)
    except Exception as err:
        log.error(err)
        return make_response(jsonify(title='Client error', description=str(err)), 400)

@app.route('/auth/bc/<auth_req_id>', methods=['GET', 'OPTIONS'])
def bc_auth_status(auth_req_id):
    if request.method == 'OPTIONS':
        return make_response('', 204)

    device = Device(
        key_dir=g.key_dir,
        config_dir=g.config_dir,
    )
    ciba = CIBAAuthentication(device)

    try:
        data = ciba.status(auth_req_id)
        return make_response(jsonify(data), 202)
    except HTTPError as err:
        if err.response.status_code == 400:
            error_message = err.response.json()['error_description']
            if error_message.startswith('authorization request is still pending'):
                return make_response(jsonify(message=error_message, status="pending"), 400)
            if error_message.startswith('backchannel authentication request is expired'):
                return make_response(jsonify(message=error_message, status="expired"), 400)
            return make_response(jsonify(message=error_message), 400)
    except Exception as err:
        log.error(err)
        return make_response(jsonify(title='Client error', description=str(err)), 400)

def main():
    '''Runs PantherX User Identity Service'''
    log.info('------')
    log.info(f"PantherX User Identity Service v{version}")
    log.info('------')

    is_superuser_or_quit()
    
    cl_arguments = get_cl_arguments()
    port = cl_arguments['port']
    key_dir = cl_arguments['key_dir']
    config_dir = cl_arguments['config_dir']
    debug = cl_arguments['debug']
    
    app.config['key_dir'] = key_dir
    app.config['config_dir'] = config_dir
    app.config['debug'] = debug
    
    if debug:
        ch.setLevel(logging.DEBUG)


    serve(app, host='127.0.0.1', port=port)

if __name__ == '__main__':
    main()

